//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

//  Files-11 Structure level 2 

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "definitions.h"
#include "f11asrt.h"
#include "devio.h"
#include "files11.h"
#include "f11liberr.h"
#include "f11file.h"
#include "f11vol.h"
#include "f11file_dir.h"

// Static Function Declarations
static int	filename_wild(const char*	file_spec);

static void	filenamescan(const char*	file_spec, 
			     dword_t*		p_length, 
			     dword_t*		p_version);

static int	filename_match(const char*	template_fname, 
			       word_t		template_fname_len,
			       const char*	fname,
			       word_t		fname_len);

void f11_directory::decompose_wcc(
	dword_t			wcc, 
	ODS_dir_record**	p_dr,
	ODS_dir_entry**		p_e,
	byte_t*			p_ct2 )
{
    dword_t		ct1;
    byte_t		ct2;

    ct1 = wcc >> 8;
    ct2 = (byte_t)(wcc & 0x000000FF);

    *p_dr = (ODS_dir_record*)((word_t*)(this->p_content_window->p_content) + ct1);
    *p_e = (*p_dr)->get_entry(ct2);
    *p_ct2 = ct2;
}

dword_t f11_directory::compose_wcc(
	ODS_dir_record*		dr,
	byte_t			ct2 )
{
    dword_t wcc;
    dword_t		ct1;

    ct1 = (word_t*)dr - (word_t*)(this->p_content_window->p_content);

    wcc = (ct1 << 8) | (ct2 & 0x000000FF);
    return wcc;
}


ODS_dir_entry*  f11_directory::f11_direntry_lookup(
	const char*		fspec, 
	word_t			wild,
	ODS_dir_record*		start_dr,
	byte_t			ct2,
	ODS_dir_record**	p_dr,
	ODS_dir_record**	p_cur_dr ) const
{
    ODS_dir_record*	dr;
    ODS_dir_record*	prev_dr;
    ODS_dir_entry*	ret = NULL;

    dword_t		length, version;
    int			has_wild = 0;

    ASSERT(fspec);

    has_wild = filename_wild(fspec);

    if (has_wild && !wild) return NULL;

    // Calculate input file_spec length and vesion
    filenamescan(fspec, &length, &version);

    // Move to the current record
    dr = start_dr;

    for (;(dr != NULL) && (ret == NULL); 
	 ct2 = 0, dr = f11_get_next_record(dr))
    {
	prev_dr = dr;

	if (dr->is_empty_record()) {
	    TRACE(("f11_direntry_lookup: empty record, go to next record\n"));
	    // Go to next record
	    continue;
	}

	if (ct2 >= dr->get_entry_count()) {
	    TRACE(("f11_direntry_lookup: cnt >= entry_count, go to next record, ct2=%02X\n", ct2));
	    // Last entry in this record has already been touched previuos time,
	    // go to next one
	    continue;
	}

	// Valid directory record
	TRACE(("f11_direntry_lookup: valid record, comparing `%s` <--> `%-.*s'\n", fspec, (int)dr->dir_b_namecount, dr->dir_t_name));

	// Is the name matching?
	if(filename_match(fspec, length, (char*)dr->dir_t_name, dr->dir_b_namecount))
	{
	    // We've found matching record... let's look for matching entry
	    if(version == 0) {
		// Return highest (i.e. first) version and move to next for next lookup
		TRACE(("f11_direntry_lookup: highest version - OK, ct2=%02X\n", ct2));
		ret = dr->get_entry(ct2);
	    } else {
		// See for specified version...
		for(byte_t i = ct2; i < dr->get_entry_count(); i++) {
		    ODS_dir_entry* e = dr->get_entry(i);
		    if (e->dir_w_version == version) {
			// Return it and move to next for next lookup
			TRACE(("f11_direntry_lookup: explicite version %d - OK, ct2=%02X\n", version, ct2));
			ret = e;
		    }
		}
	    }
	}
    }

    if(ret == NULL) {
	TRACE(("f11_direntry_lookup: FILE_NOT_FOUND, return\n======\n"));
	dev->set_error(F11_FILE_NOT_FOUND);
	return NULL;
    }

    *p_dr = dr;
    if (p_cur_dr != NULL)
	*p_cur_dr = prev_dr;
    return ret;
}

ODS_dir_record* f11_directory::f11_get_next_record(ODS_dir_record* dr) const
{
    dword_t 	blk = 0;

    // Is current record 'end of records' marker?
    if (dr->dir_w_size == 0xFFFF) {

	// Next record is in the next dir block or not present
	blk = ((size_t)dr - (size_t)(this->p_content_window->p_content)) / F11_BLOCK_SIZE + 1;
	blk *= F11_BLOCK_SIZE;
	if (blk >= get_fhdr()->get_file_length()) {

	    // No more blocks
	    dev->set_error(F11_FILE_NOT_FOUND);
	    return 0;
	}

	dr = (ODS_dir_record*)((byte_t*)(this->p_content_window->p_content) + blk);

    } else {

	// Next record is in the current block
	dr = (ODS_dir_record*)((byte_t*)dr + dr->dir_w_size + 2);
    }

    return dr;
}

void f11_directory::remove_dir_entry(
			ODS_dir_record*	    dr,
			ODS_dir_entry*	    e )
{
    byte_t*	buffer;

    word_t ent_num;
    ent_num = (dr->dir_w_size - sizeof(ODS_dir_record)
           - dr->dir_b_namecount - (dr->dir_b_namecount & 1) + 6 - 2) / sizeof(ODS_dir_entry);
    TRACE(("remove_dir_entry: %-.*s;%d, entry num = %d\n", dr->dir_b_namecount, dr->dir_t_name, e->dir_w_version, ent_num));

    // Remove the entry from the record that contains more than 1 entry.
    // In this case we may remove only one entry, and not  the whole (empty) dir record.
    buffer = (byte_t*)dr - (((byte_t*)dr - (byte_t*)p_content_window->p_content) % F11_BLOCK_SIZE);
    byte_t* ne = (byte_t *)e + sizeof(ODS_dir_entry);
    word_t e_remainder = 512 - (ne - buffer);
    ASSERT((e_remainder >= 0) || (e_remainder < 512));
    memcpy(e, ne, e_remainder);
    dr->dir_w_size = dr->dir_w_size - sizeof(ODS_dir_entry);
    flag_content_modified = 1;
    
    // Add indicator of the skipping record
    *(word_t*)((byte_t*)e + e_remainder) = -1;

    if (dr->get_entry_count() == 0)
    {
	// The dir record is empty now, remove it as well...
	TRACE(("remove_dir_entry: removing empty dir record...\n"));
	byte_t* nr = (byte_t *)dr + dr->dir_w_size + 2;
	word_t	r_remainder = 512 - (nr - buffer);
	ASSERT((r_remainder >= 0) || (r_remainder < 512));
	memcpy(dr, nr, r_remainder);

	// Add indicator of the skipping record
	*(word_t*)((byte_t*)dr + r_remainder) = -1;

	if (((byte_t*)dr == buffer) && dr->is_empty_record())
	{
	    // If the whole block is empty, shrink the directory file contents
	    dword_t eofblk = get_fhdr()->get_eofblk();
	    
	    ASSERT(get_fhdr()->get_ffbyte() == 0);
	    TRACE(("remove_dir_entry: shrinking file size by 1 block...\n"));

	    for(dword_t nextblock = ((byte_t*)buffer - (byte_t*)p_content_window->p_content) / F11_BLOCK_SIZE + 1; 
		nextblock < eofblk - 1;
		nextblock++) 
	    {
		memcpy((nextblock-1)*F11_BLOCK_SIZE+(byte_t*)p_content_window->p_content, 
			nextblock*F11_BLOCK_SIZE+(byte_t*)p_content_window->p_content, 
			F11_BLOCK_SIZE);
	    }

	    get_fhdr()->set_eofblk(eofblk - 1);
	    flag_header_modified = 1;
	}
    }
    this->f11_sync();
}

// Static functions

static int filename_wild(const char* file_spec)
{
    char*	b1;
    char*	b2;
    int wild	= 0;

    b1 = strchr(file_spec, '*');
    b2 = strchr(file_spec, '%');

    if ((b1 != 0) || (b2 != 0)) {
	wild = 1;
    }
    return wild;
}

static void filenamescan(const char* file_spec, dword_t* p_length, dword_t* p_version)
{
    char*	b;
    char*	p_e;
	
    b = strchr(file_spec, ';');
    if(b != NULL) {
	*p_length  = (dword_t)(b - file_spec);
	*p_version = (dword_t)strtol(b + 1, &p_e, 10);
    } else {
	*p_length  = (dword_t)strlen(file_spec);
    }
}

static int filename_match(
		const char*	template_fname, 
		word_t		template_fname_len,
		const char*	fname,
		word_t		fname_len)
{

#if defined(NO_WILDS)
    // This version does not match wilds
    return (fname_len == template_fname_len &&
	   strncmp(template_fname, fname, fname_len) == 0);
#else
    int percent = 0;
    const char *spc = template_fname, *ent = fname;
    const char *spcend = spc + template_fname_len, *entend = ent + fname_len;

    // See how much matches without wildcards...
    if (spc < spcend && ent < entend) {
        int count = template_fname_len;
        if (fname_len < count) 
	    count = fname_len;
        do {
            char sch = *spc, ech = *ent;
            if (sch != ech) 
		if (toupper(sch) != toupper(ech))
                    if (sch == '%') {
                        percent = 1;
                    } else {
                        break;
                    }
            spc++;
            ent++;
        } while (--count > 0);
    }

    TRACE(("Namematch %-.*s [%d] %-.*s [%d]\n", 
	template_fname_len, template_fname, template_fname_len, 
	fname_len, fname, fname_len));

    // Mismatch - return result unless wildcard...
    if (spc >= spcend) {
        if (ent >= entend) {
            return 1;	    //EQ
        } else {
            if (percent) {
                return 0;   //NE
            } else {
                return 0;   //GT  - Entry longer than search spec
            }
        }
    } else {
        int offset = 0;
        if (*spc != '*') {
            if (percent) 
		return 0;	//NE
            if (ent < entend)
                if (toupper(*ent) > toupper(*spc)) 
		    return 0;	//GT
            return 0;		//LT
        }

        // See if we can find a match with wildcards
        spc++;
        if (spc < spcend) {
            do {
                if (spc >= spcend) {
                    if (ent >= entend) 
			break;
                    spc -= offset;
                    ent -= offset - 1;
                    offset = 0;
                } else {
                    char sch = toupper(*spc);
                    if (sch == '*') {
                        offset = 0;
                        spc++;
                        if (spc >= spcend) 
			    break;
                    } else {
                        if (ent < entend) {
                            char ech = toupper(*ent);
                            if (sch == ech || sch == '%') {
                                offset++;
                                spc++;
                                ent++;
                            } else {
                                spc -= offset;
                                ent -= offset - 1;
                                offset = 0;
                            }
                        } else {
                            return 0; //NE
                        }
                    }
                }
            } while (1);
        }
    }
    return 1; //EQ

#endif
}

